from pandac.PandaModules import *
from random import random
import textwrap

class Building:
    model = None
    cost = 100
    upkeep = 1
    power = 1
    water = 0
    
    #tmp
    radius = .2
    
    drain = {}
    pour = {}
    
    order = 0
    
    size = 1
    
    safe_zone = None
    system = False
    
    houses = 0
    
    def load(self):
        name = self.__class__.__name__.lower()
        if name in game.resources.__dict__:
            
            self.node = game.resources.__dict__[name].copyTo(render)
            #b = self.node.getBounds()
            #r = (b.getMin() - b.getMax()).length()
            #s = 1/r*5  
            #print "scaling",s
            self.node.setScale(self.size*5)
            return self.node
        else:
            print "no",repr(name)," in game"
        
    def simulate(self):
        
        if self.full_check():
            for type,amount in self.drain.iteritems():
                game.rs.drain(self,type,amount)
                
            if self.power > 0 :
                game.ps.drain(self,'power',self.power)
            elif self.power < 0 :
                game.ps.pour(self,'power',-self.power)
            else:
                pass # power not needed
                
            if self.water > 0 :
                game.ws.drain(self,'water',self.water)
            elif self.water < 0:
                game.ws.pour(self,'water',-self.water)
            else:
                pass # water not needed
            
            if self.upkeep > 0 :
                game.rv.upkeep(self.upkeep)
            elif self.upkeep < 0:
                game.rv.commurce(-self.upkeep)
            else:
                pass # money not needed

            game.rv.power(self.power)
            game.rv.water(self.water)
            game.rv.houses(self.houses)
                
            for type,amount in self.pour.iteritems():
                game.rs.pour(self,type,amount)
            self.node.setColor(1,1,1)
        else:
            self.node.setColor(1,0,0)
            game.ms.post(str(self)+" is inactive because of "+self.reason,-100)        
        
    def after_simulate(self):
        pass
        
    def full_check(self):
        for type,amount in self.drain.iteritems():
            if not game.rs.check(self,type,amount):
                self.reason = "not enough "+type
                return False
        if self.power > 0 and not game.ps.check(self,'power',self.power):
            self.reason = "not enough power "+str(self.power)
            return False
        if self.water > 0 and not game.ws.check(self,'water',self.water):
            self.reason = "not enough water "+str(self.water)
            return False
        if self.upkeep > 0 and not game.rv.check(self.upkeep):
            self.reason = "not enough money $"+str(self.upkeep)
            return False
        return True
    
    def can_build_check(self,pos):
        print "at",pos.getZ()
        if pos.getZ() < game.sea_level:
            game.ms.post(str(self)+" can't be build bellow sea level")
            return False
        
        if game.ps.too_close(self,pos):
            return False
            
        
        return True 
  
    def __str__(self):
        return self.__class__.__name__
   
    @classmethod
    def help(self):
        if self.__doc__:
            text = "\n".join(textwrap.wrap(" ".join(self.__doc__.split()), 30))
            
        else:
            text = ""
        header = [self.__name__,':\n\n']
        if self.cost: header.extend(['cost:',str(self.cost),'\n'])
        if self.upkeep > 0: header.extend(['upkeep:',str(self.upkeep),'\n'])
        if self.upkeep < 0: header.extend(['earns:',str(-self.upkeep),'\n'])
        if self.power > 0: header.extend(['power:',str(self.power),'\n'])
        if self.water > 0: header.extend(['water:',str(self.water),'\n'])
        if self.power < 0: header.extend(['produce power:',str(-self.power),'\n'])
        if self.water < 0: header.extend(['filters water:',str(-self.water),'\n'])
        if self.drain: header.extend(['use:',repr(self.drain),'\n'])
        if self.pour: header.extend(['produce:',repr(self.pour),'\n'])
        header.extend(['\n',text])
        return ''.join(header)
   
   
class Command(Building):
    """
        Command center is use to jump start
        a new city or recovering city 
        it generates a small supply of workers,
        power and water.
        
        Build roads, tubes and power lines to it.
        
        You can't have more then one    
    """
    cost = 0
    upkeep = -1
    power = -1
    water = -20
    pour = {'workers':1}
    houses = 100
    size = .1
    
    def can_build_check(self,pos):
        for b in game.bs.buildings:
            if b.__class__ == Command:
                 game.ms.post(str(self)+" can't have more then one command center")
                 return False
        return Building.can_build_check(self,pos)
     
class Residential(Building):
    """
        Is a standard building to house your 
        populations.  It requires a steady 
        supply of food and water in exchange
        it houses your population and produces 
        workers to keep all the machines in the
        city running    
    """
    cost = 260
    upkeep = 4
    power = 5
    water = 6
    drain = {'food':100}
    pour = {'workers':1000}
    
    houses = 1000
    size = .0006
        
class ColonyResidential(Building):
    """
        Simple self powering housing system 
    """
    cost = 60
    upkeep = 1
    power = 0
    pour = {'workers':10}
    houses = 10
    size = .03  
       
class SmallResidential(Building):
    """
        Simple self powering housing system requires food
    """
    cost = 160
    upkeep = 2
    power = 0
    drain = {'food':10}
    pour = {'workers':100}
    houses = 100
    size = .0003
       
class Farm(Building):
    
    """
        The farming robots requires just 
        a few people supervise the process
        of using mainly water to grow food. 
    
    """
    cost = 150
    upkeep = 2
    power = 5
    water = 10
    drain = {'workers':20}
    pour = {'food':100}
    size = .00008
    
class ColonyFarm(Building):    

    """
        The farming robots making food for the human population.
        It requires just a few workers to run (build a road from your 
        command center)    
    """
    cost = 50
    upkeep = .5
    power = 0
    pour = {'food':5}
    drain = {'workers':1}
    size = .00004
    
class SmallFarm(Building):    

    """
        The farming robots making food for the human population.
        It requires just a few workers to run (build a road from your 
        command center)    
    """
    cost = 100
    upkeep = 1
    power = 0
    water = 5
    pour = {'food':15}
    drain = {'workers':3}
    size = .00006
    
class Aqua(Building):
    
    """
        In order to get water to be used in 
        your city you have to purify it and
        send it to your buildings.
        
        This structure is build bellow 
        sea-level and requires under water 
        power lines to connected it to the 
        power gird.
    
    """
    cost = 150
    upkeep = 10
    power = 5
    #drain = {'workers':1}
    water = -40
    size = .13
        
    def can_build_check(self,pos):
        print "at",pos.getZ()
        if pos.getZ() >= game.sea_level:
            game.ms.post(str(self)+" needs to be at or bellow sea level")
            return False
        return True
    
class ColonyAqua(Aqua):    
    cost = 50
    upkeep = 10
    power = 0
    water = -20
    size = .1
    
class Factory1(Building):
    """
        Using robots the factories can produce 
        a very large amount of good to be sent 
        via space ports to other planets or sold
        here at the markets.
        
    """
    cost = 100
    upkeep = 1
    power = 1
    drain = {} #{'workers':1,
            #'resources':10}
    pour = {'goods':10}
       
    size = .03
    
class Factory2(Building):
    
    cost = 500
    upkeep = 5
    power = 25
    drain = {'workers':8,
            'resources':80}
    pour = {'goods':80}
    size = .04
     
class Factory3(Building):
    
    cost = 1000
    upkeep = 10
    power = 50
    drain = {'workers':10,
            'resources':100}
    pour = {'goods':120}
    size = .06
     
class Factory4(Building):
    
    cost = 2000
    upkeep = 1
    power = 5
    drain = {'workers':80,
            'resources':370}
    pour = {'goods':400}         
    size = .07
         
class Factory5(Building):
    
    cost = 1000
    upkeep = 100
    power = 500
    drain = {'workers':100,
            'resources':1000}
    pour = {'goods':1000}
    size = .1
    
class Reactor1(Building):
    """
        Reactors take water and using the water
        they create power.  Power is needed for almost 
        all other buildings.  Unfortunately  the radiation
        laws require other buildings to be build some 
        distance form the reactors. 
    
    """
    
    cost = 250
    upkeep = 1
    power = -50
    water = 1
    size = .05
    safe_zone = 16
        
class Reactor2(Building):
    
    cost = 100
    upkeep = 1
    power = 5
    drain = {'workers':1,
            'resources':10}
    pour = {'goods':10}
                
        
        
class Mine(Building):
    
    cost = 100
    upkeep = 1
    power = 5
    drain = {'workers':1}
    pour = {'goods':10}
                
    size = .1
        
class Market(Building):
    """
        Move your goods here so that they can be 
        sold for profit    
    """
    cost = 100
    upkeep = -100
    power = 5
    drain = {'workers':1,'goods':10}
    size = .0001
    
class ColonyMarket(Market):
    """
        This is a small market used to sell food   
    """
    cost = 100
    upkeep = -10
    power = 0
    drain = {'workers':10,'food':1}

class StarPort(Building):
    """
        Move your goods here so that they can be 
        sold for profit to other cities on other planets    
    """
    
    cost = 200
    upkeep = -10
    power = 0
    size = .00025
    drain = {'goods':1}
    
class Storage(Building):
    
    
    cost = 100
    upkeep = 1
    power = 5
    drain = {'goods':1,'goods':10}
    
    def simulate(self):
        # drain as much as you can if the road 
        # is getting full with goods else 
        # pour goods out 
        pass
    
class Power(Building):
    
    cost = 250
    upkeep = 1
    generate = 5
    drain = {'workers':1}
    
    def simulate(self):
        # drain as much as you can and check for local demand
        if self.full_check():
            for type,amount in self.drain.iteritems():
                game.rs.darin(type,amount)
            #give power
            game.ps.add(self.generate)
            
class RoadNode(Building):
    
    """
        Road nodes can transport people goods or resources
        most buildings require some roads leading to them
    """
    
    cost = 10
    upkeep = .1
    generate = 5
    speed = 100
    order = 100
    
    size = .3
    
    store = 100
    
    def __init__(self):
        self.connected = []
        self.resources = {}
    
    def simulate(self):
       
        divider = len(self.connected)
        
        for res_type in self.resources:
            low_amount = 1000000
            low_node = None
            for node in self.connected:
                amount = node.check(res_type)
                if low_amount > amount:
                    low_amount = amount
                    low_node = node
                     
            if low_node and low_amount < self.resources[res_type]:
                amount = min(self.speed*random(),self.resources[res_type]*.5) #min(self.speed,self.resources[res_type]/divider)
                low_node.resources[res_type] += amount
                self.resources[res_type] -= amount
        
    def debug(self):
        if self not in game.rs.roads:
            print "*",self,"is a lone road"
        
    def read_out(self):
        readout = []
        for res_type in self.resources:
            if self.resources[res_type] > 0:
                readout.append("%s:%i" %(res_type,self.resources[res_type]))
        self.text.setText("\n".join(readout))
            
    def load(self):
        node = Building.load(self)
        self.text = TextNode("text")            
        self.text.setCardDecal(True)
        self.text.setUsageHint(GeomEnums.UHDynamic)
        self.text.setWordwrap(50)
        self.text.setText("loading python...")
        textNode = NodePath("text")           
        textNode.attachNewNode(self.text)     
        textNode.setScale(.1*1/self.size)
        textNode.setBillboardPointEye()
        textNode.reparentTo(node)
        textNode.setPos(0,0,2)
        textNode.setLightOff(True)
        return node
    
    def check(self,type):
        if type not in self.resources:
            self.resources[type]  = 0
        return self.resources[type] 
    
    def drain(self,type,ammount):
        if type not in self.resources:
            print "over drain"
        else:
            self.resources[type] -= ammount
            if self.resources[type] < 0 :
                print "over drain"
            
    def pour(self,type,ammount):
        if type not in self.resources:
            self.resources[type] = ammount
        else:
            self.resources[type] += ammount
    
    def after_simulate(self):
        for res_type in self.resources:
             if self.resources[res_type] > self.store:
                 self.resources[res_type] = self.store
       
        if game.selected == self.__class__: 
            self.read_out()
        else:
            self.text.setText("")
        self.debug() 
         
        return 
        # PUT THIS SYSTEM BACK:
        fill = 0
        for res_type in self.resources:
             fill += self.resources[res_type]
        
        if fill > self.store:
            surplus = self.store/fill
            for res_type in self.resources:
                self.resources[res_type] *= surplus
        
        
    
class WaterNode(RoadNode):
    
    """
        Fewer building require water access.
        Water starts its cleaning process at sea-level
        with the auqa filters and then is required by
        reactors, farms and residential complexes.
     
    """
    
    cost = 10
    upkeep = .1
    generate = 5
    size = .1
    
    def debug(self):
        if self not in game.ws.roads:
            print "*",self,"is a lone road"
        
    def read_out(self):
        readout = []
        for res_type in self.resources:
            if self.resources[res_type] > 0:
                readout.append("%s:%i tons" %(res_type,self.resources[res_type]))
        self.text.setText("\n".join(readout))
    
class EnergyNode(RoadNode):
    
    """
        Most buildings require power.  
        So build energy nodes every where!    
    """
    
    cost = 10
    upkeep = .1
    generate = 5
    speed = 1000
    
    def debug(self):
        if self not in game.ps.roads:
            print "*",self,"is a lone road"
        
    def read_out(self):
        readout = []
        for res_type in self.resources:
            if self.resources[res_type] > 0:
                readout.append("%s:%i mw" %(res_type,self.resources[res_type]))
        self.text.setText("\n".join(readout))
    
    