from direct.task import Task
from direct.showbase.DirectObject import DirectObject
import direct.directbase.DirectStart
from pandac.PandaModules import *
import math

class EggOctree(object):

   def build(self,vertexPool,group, destNode, cellSize, collide=False):
      snode = group
      snode.triangulatePolygons(0xff)
      # First, extract the <VertexPool>
      #vpool = snode.getFirstChild()
      #assert type(vpool) == EggVertexPool, 'First child must be a <VertexPool>'
      vpool = vertexPool
      
      # Loop on the vertices determine the bounding box
      minBB = Point3D()
      maxBB = Point3D()
      nverts = vpool.getHighestIndex()
      for iv in range(nverts + 1) :
         vtx = vpool[iv]
         vtxPos = vtx.getPos3()
         # Sets the X, Y or Z components
         i = 0
         for set in [Point3D.setX, Point3D.setY, Point3D.setZ]  :
            if vtxPos[i] < minBB[i] :
               set(minBB, vtxPos[i])
            if vtxPos[i] > maxBB[i] :
               set(maxBB, vtxPos[i])
            i += 1
      minBB -= Vec3D(0.001, 0.001, 0.001)
      maxBB += Vec3D(0.001, 0.001, 0.001)
      # Number of leaves x,y,z
      bboxSize = maxBB - minBB
      self.ncx = math.ceil(bboxSize.getX() / cellSize.getX())
      self.ncy = math.ceil(bboxSize.getY() / cellSize.getY())
      self.ncz = math.ceil(bboxSize.getZ() / cellSize.getZ())
      # Depth of the tree x,y,z
      self.depthX = math.ceil(math.log(self.ncx) / math.log(2))
      self.depthY = math.ceil(math.log(self.ncy) / math.log(2))
      self.depthZ = math.ceil(math.log(self.ncz) / math.log(2))
      self.depth = max(self.depthX, self.depthY, self.depthZ)
      self.cells = [[[
            {'min':
               Point3D(minBB.getX() + x * cellSize.getX(),
                     minBB.getY() + y * cellSize.getY(),
                     minBB.getZ() + z * cellSize.getZ()),
            'max':
               Point3D(minBB.getX() + (x+1) * cellSize.getX(),
                     minBB.getY() + (y+1) * cellSize.getY(),
                     minBB.getZ() + (z+1) * cellSize.getZ()),
            'group':EggGroup('leaf_%d_%d_%d' % (x,y,z)) }
            for z in range(int(self.ncz))]
            for y in range(int(self.ncy))]
            for x in range(int(self.ncx))]
      print 'Cell grid is %dx%dx%d' % (len(self.cells), len(self.cells[0]), len(self.cells[0][0]))
      if collide :
         for x in range(int(self.ncx)) :
            for y in range(int(self.ncy)) :
               for z in range(int(self.ncz)) :
                  self.cells[x][y][z]['group'].addObjectType('barrier')
      
      print "Iterate on the <Polygon>s (should be triangles)"
      poly = snode.getFirstChild()
      print "h?"
      while poly != None :
         if isinstance(poly, EggPolygon) :
            #print "Get the triangle center"
            polyCenter = Point3D(0,0,0)
            for i in range(3) :
               polyCenter += poly.getVertex(i).getPos3()
            polyCenter /= 3.0
            # Add the triangle to the corresponding cell group
            cx = int(math.floor((polyCenter.getX()-minBB.getX()) / cellSize.getX()))
            cy = int(math.floor((polyCenter.getY()-minBB.getY()) / cellSize.getY()))
            cz = int(math.floor((polyCenter.getZ()-minBB.getZ()) / cellSize.getZ()))
            self.cells[cx][cy][cz]['group'].addChild(poly)
         poly = snode.getNextChild()
      
      print "# Add the vertex data"
      destNode.addChild(vpool)
      self.nleaves = self.recur(destNode, 0, 0,0,0)
      print self.nleaves, 'leaves added'

   def recur(self, node, depth, x, y, z) :
      if depth < self.depth :
         nnode = EggGroup('')
         delt = int(math.pow(2, self.depth - depth - 1))
         nchilds = 0
         nchilds += self.recur(nnode, depth+1, x, y, z)
         nchilds += self.recur(nnode, depth+1, x + delt, y, z)
         nchilds += self.recur(nnode, depth+1, x, y + delt, z)      
         nchilds += self.recur(nnode, depth+1, x + delt, y + delt, z)
         nchilds += self.recur(nnode, depth+1, x, y, z + delt)
         nchilds += self.recur(nnode, depth+1, x + delt, y, z + delt)
         nchilds += self.recur(nnode, depth+1, x, y + delt, z + delt)
         nchilds += self.recur(nnode, depth+1, x + delt, y + delt, z + delt)
         if nchilds > 0 :
            node.addChild(nnode)
         return nchilds
      else :
         # We are in a leaf
         if x < self.ncx and y < self.ncy and z < self.ncz :
            node.addChild(self.cells[x][y][z]['group'])
            return 1
         else :
            return 0 


def getChildren(eggNode):
    try:
        child = eggNode.getFirstChild()
        while child:
            yield child
            child = eggNode.getNextChild()
    except:
        pass

def eggLs(eggNode,indent=0):
    if eggNode.__class__.__name__ != "EggPolygon":
        print " "*indent+eggNode.__class__.__name__+" "+eggNode.getName()
        for eggChildren in getChildren(eggNode):
            eggLs(eggChildren,indent+1)
        
        
        
def eggStripTexture(eggNode):
    if eggNode.__class__ == EggPolygon:
        eggNode.clearTexture()        
    else:
        for eggChildren in getChildren(eggNode):
            eggStripTexture(eggChildren)
            
            
infile = "map.egg"
outfile = "mapc.egg"
         
egg = EggData()
egg.read(Filename(infile))
eggLs(egg)
      
sourceNode = egg.getFirstChild()
while sourceNode:
    print type(sourceNode)
    if type(sourceNode) == EggVertexPool:
        vertexPool = sourceNode
    if type(sourceNode) == EggGroup:
        group = sourceNode # the sourcenode must contain a VertexPool and a list of Polygons
    
    sourceNode = egg.getNextChild() # the sourcenode must contain a VertexPool and a list of Polygons

if vertexPool and group:
    
    ed = EggData()
    ed.setCoordinateSystem(egg.getCoordinateSystem())
    octree = EggOctree()
    octree.build(vertexPool,group, ed, Vec3(2.50, 2.50, 2.50), collide=True)
    #eggLs(ed)
    #group.addChild(vertexPool)
    #eggLs(ed)
    eggStripTexture(ed)
    #print >> open(outfile,'w'), ed
    ed.writeEgg(Filename(outfile))


#np = loader.loadModel(outfile)
#np.ls()
#np.analyze()


